import unittest

from fss_utils.sshkey import FABRICSSHKey, FABRICSSHKeyException


class FABRICSSHKeyTest(unittest.TestCase):

    def setUp(self) -> None:
        pass

    def tearDown(self) -> None:
        pass

    def testGenerateRSA(self):
        """
        Test generation of various key types
        :return:
        """
        rsa = FABRICSSHKey.generate("rsa_key@localhost", "rsa")
        rsa.get_fingerprint()
        self.assertTrue(rsa.length >= 3072)

        priv, pub = rsa.as_keypair()

        rsa1 = FABRICSSHKey(pub)
        self.assertEqual(rsa1.length, rsa.length)
        self.assertEqual(rsa1.get_fingerprint(), rsa.get_fingerprint())

        with open('rsa_key.priv', 'w') as f:
            f.write(priv)

        with open('rsa_key.pub', 'w') as f:
            f.write(pub)

    def testGenerateECDSA(self):
        """
        Test generation of various key types
        :return:
        """
        ecdsa = FABRICSSHKey.generate("ecdsa_key@localhost", "ecdsa")
        ecdsa.get_fingerprint()
        # We don't validate ECDSA key length - sometimes it is 249 or other numbers
        # self.assertTrue(ecdsa.length >= 255)

        priv, pub = ecdsa.as_keypair()

        ecdsa1 = FABRICSSHKey(pub)
        self.assertEqual(ecdsa1.length, ecdsa.length)
        self.assertEqual(ecdsa1.get_fingerprint(), ecdsa.get_fingerprint())

        with open('ecdsa_key.priv', 'w') as f:
            f.write(priv)

        with open('ecdsa_key.pub', 'w') as f:
            f.write(pub)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2021 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author Ilya Baldin (ibaldin@renci.org)

from http import HTTPStatus
from flask import Response, request


class HTTPErrorTuple:
    """
    Small helper class to return HTTP error in the format UIS, PR like.
    Use it as follows:
    return HTTPErrorTuple(status=HTTPStatus.INTERNAL_SERVER_ERROR,
                          body="Non standard error return",
                          xerror="Horrible internal error details").astuple()
    You will get a flask.Response
    ("Non standard error return", 500, { "X-Error": "Horrible internal error details"})
    If you don't want CORS headers included, add cors=False as a parameter to constructor.
    """

    def __init__(self, *, request, status, body: str = None, xerror: str = None, cors: bool = True):
        self.request = request
        self.cors = cors
        self.body = body if body is not None else status.phrase
        self.code = status.real
        self.xerror = xerror

    def response(self) -> Response:
        r = Response()
        r.status_code = self.code
        r.data = self.body
        if self.cors:
            r.headers['Access-Control-Allow-Origin'] = self.request.headers.get('Origin', '*')
            r.headers['Access-Control-Allow-Credentials'] = 'true'
            r.headers['Access-Control-Allow-Methods'] = 'GET, POST, PUT, DELETE, OPTIONS'
            r.headers['Access-Control-Allow-Headers'] = 'DNT, User-Agent, X-Requested-With, ' \
                                                        'If-Modified-Since, Cache-Control, ' \
                                                        'Content-Type, Range'
            r.headers['Access-Control-Expose-Headers'] = 'Content-Length, Content-Range, X-Error'

        if self.xerror:
            r.headers['X-Error'] = self.xerror

        return r


def cors_response(status, body: str = None, xerror: str = None, cors: bool = True) -> Response:
    """
    Shortcut method to provide a cors response
    :param status:
    :param body:
    :param xerror:
    :param cors:
    :return:
    """
    r = HTTPErrorTuple(request=request, status=status, body=body, xerror=xerror, cors=cors)
    return r.response()



import setuptools
from fss_utils import __VERSION__

with open("README.md", "r") as fh:
  long_description = fh.read()

with open("requirements.txt", "r") as fh:
  requirements = fh.read()

setuptools.setup(
  name="fabric_fss_utils",
  version=__VERSION__,
  author="Ilya Baldin",
  description="FABRIC System Services Utilities library",
  url="https://github.com/fabric-testbed/system-service-utils",
  long_description="FABRIC System Services Utilities library",
  long_description_content_type="text/plain",
  packages=setuptools.find_packages(),
  include_package_data=True,
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  python_requires=">=3.7",
  install_requires=requirements,
  setup_requires=requirements,
)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org) Michael Stealey (stealey@renci.org)

from typing import Tuple
import re
import base64
import hashlib

from cryptography.hazmat.primitives.asymmetric import rsa, ec
from cryptography.hazmat.primitives import serialization

# this is how we set what key type is produced by default and default key length
KEY_ALGORITHMS = {
    "rsa": (rsa.generate_private_key(public_exponent=65537, key_size=3072), 3072),
    "ecdsa": (ec.generate_private_key(ec.SECP256R1()), 256)
}

LABEL_REGEX = "^[a-zA-Z0-9@-_()]{0,64}$"


class FABRICSSHKey:

    def __init__(self, public_key: str):
        """
        Load the public portion of the key
        """
        assert public_key is not None
        # this also validates that this is either DSA or ECDSA key
        self._length = FABRICSSHKey.get_key_length(public_key, validate=True)
        self._name, self._public_key, self._label = public_key.split(" ")
        # can only be generated
        self._private_key = None

    @property
    def name(self):
        return self._name

    @property
    def public_key(self):
        return self._public_key

    @property
    def label(self):
        return self._label

    @property
    def length(self):
        return self._length

    @property
    def private_key(self):
        return self._private_key

    @classmethod
    def generate(cls, label: str, algorithm: str):
        assert label is not None
        matches = re.match(LABEL_REGEX, label)
        if matches is None:
            raise FABRICSSHKeyException(f'Label {label} does not match expected regular expression {LABEL_REGEX}')
        # generate a key
        if algorithm not in KEY_ALGORITHMS.keys():
            raise FABRICSSHKeyException(f'Key Algorithm configured as {algorithm} is not supported')
        # this calls to generate
        key = KEY_ALGORITHMS[algorithm][0]

        # get the private key in PEM format
        private_key = key.private_bytes(encoding=serialization.Encoding.PEM,
                                        format=serialization.PrivateFormat.OpenSSH,
                                        encryption_algorithm=serialization.NoEncryption()).decode('utf-8')
        public_key_with_name = key.public_key().public_bytes(encoding=serialization.Encoding.OpenSSH,
                                                             format=serialization.PublicFormat.OpenSSH).decode('utf-8')

        # just append the label
        ret = FABRICSSHKey(" ".join([public_key_with_name.strip(),
                                     label]))
        ret._private_key = private_key
        return ret

    def as_keypair(self) -> Tuple[str, str]:
        """
        Returns a typle of private, public keys formatted for SSH
        :return:
        """
        return self.private_key, self.as_public_key_string()

    def as_public_key_string(self):
        return " ".join([self._name, self._public_key, self._label])

    def get_fingerprint(self) -> str:
        """
        Generate MD5 fingerprint. Use 'ssh-keygen -lf <filename> -E md5' to get the same output
        """
        # for some reason Paramiko fingerprint() does not match the output of sshkeygen -lf <key> -E md5
        # so we are rolling our own based on
        # https://stackoverflow.com/questions/64733471/how-to-calculate-a-fingerprint-from-an-rsa-public-key
        rawdata = base64.b64decode(self._public_key)
        hexdigest = hashlib.md5(rawdata).hexdigest()
        keychunks = [hexdigest[i:i+2] for i in range(0, len(hexdigest), 2)]
        return 'MD5:' + ":".join(keychunks)

    @staticmethod
    def get_key_length(ks: str, validate=False) -> int:
        """
        Returns key length of the ECC or DSA public key using horrible non-portable hacks.
        Validates the key format first by loading it as DSA or ECDSA.
        """
        try:
            ck = serialization.load_ssh_public_key(ks.encode('utf-8'))
        except:
            raise FABRICSSHKeyException(f'Provided public key starting with {ks[0:50]} cannot be imported')

        return ck.key_size

    def __str__(self):
        return f"Private:\n{self.private_key}\nPublic:\n{' '.join([self._name, self._public_key, self._label])}"

    def __repr__(self):
        return str(self)


class FABRICSSHKeyException(Exception):
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"SSH Key exception: {msg}")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author Komal Thareja (kthare10@renci.org)
import base64
import enum
import gzip
import traceback
from datetime import timedelta, datetime
from typing import Tuple, Union

import jwt
from authlib.jose import jwk
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization


@enum.unique
class ValidateCode(enum.Enum):
    VALID = 1
    UNSPECIFIED_KEY = 2
    UNSPECIFIED_ALG = 3
    UNKNOWN_KEY = 4
    INVALID = 5
    UNABLE_TO_FETCH_KEYS = 6
    UNPARSABLE_TOKEN = 7
    UNABLE_TO_DECODE_KEYS = 8
    UNABLE_TO_LOAD_KEYS = 9
    UNABLE_TO_COMPRESS = 10
    UNABLE_TO_DECOMPRESS = 11
    UNSPECIFIED_ID_TOKEN = 12

    def interpret(self, exception=None):
        interpretations = {
            1: "Token is valid",
            2: "Token does not specify key ID",
            3: "Token does not specify algorithm",
            4: "Unable to find public key at JWK endpoint",
            5: "Token signature is invalid",
            6: "Unable to fetch keys from the endpoint",
            7: "Unable to parse token",
            8: "Unable to decode public keys",
            9: "Unable to load key from file",
            10: "Unable to compress the encoded token",
            11: "Unable to decompress the encoded token",
            12: "Identity Token or Identity Claims not specified"
          }
        if exception is None:
            return interpretations[self.value]
        else:
            return str(exception) + ". " + interpretations[self.value]


class JWTManager:
    @staticmethod
    def encode_and_sign_with_private_key(*, validity: int, claims: dict, private_key_file_name: str, kid: str,
                                         algorithm: str,
                                         pass_phrase: str = None) -> Tuple[ValidateCode, Union[Exception, str]]:
        """
        sign and base64 encode the token
        :param validity validity in seconds
        :param claims claims to be added in token
        :param private_key_file_name private key file name used to sign the token
        :param kid kid to be added to the header
        :param algorithm algorithm by which to encode the token
        :param pass_phrase private key pass phrase

        :return ValidateCode or None, exception or None, Encoded Token or None:
        """
        if pass_phrase is not None and pass_phrase != "":
            pass_phrase = pass_phrase.encode("utf-8")
        else:
            pass_phrase = None

        try:
            with open(private_key_file_name) as private_key_fh:
                pem_data = private_key_fh.read()
                private_key_fh.close()
                private_key = serialization.load_pem_private_key(data=pem_data.encode("utf-8"),
                                                                 password=pass_phrase,
                                                                 backend=default_backend())
        except Exception as e:
            return ValidateCode.UNABLE_TO_LOAD_KEYS, e

        claims['iat'] = int(datetime.now().timestamp())
        claims['exp'] = int((datetime.now() + timedelta(seconds=int(validity))).timestamp())

        try:
            encoded_token = str(jwt.encode(claims, private_key, algorithm=algorithm, headers={'kid': kid}), 'utf-8')
            return ValidateCode.VALID, encoded_token
        except Exception as e:
            return ValidateCode.INVALID, e

    @staticmethod
    def encode_jwk(*, key_file_name: str, kid: str, alg: str) -> Tuple[ValidateCode, Union[Exception, dict]]:
        """
        Encode JWK from a PEM file
        :param key_file_name Key File Name
        :param kid kid to be added to the JWK
        :param alg algorithm to be added

        :return ValidateCode or None, exception or None, JWK or None:
        """
        pem_data = None
        try:
            with open(key_file_name) as public_key_fh:
                pem_data = public_key_fh.read()
                public_key_fh.close()
        except Exception as e:
            return ValidateCode.UNABLE_TO_LOAD_KEYS, e

        try:
            result = jwk.dumps(pem_data, kty='RSA')
            result["kid"] = kid
            result["alg"] = alg
            result["use"] = "sig"
            return ValidateCode.VALID, result
        except Exception as e:
            return ValidateCode.INVALID, e

    @staticmethod
    def encode_and_compress(*, claims: dict, secret: str, validity: int, algorithm: str = 'HS256',
                            compression: bool = True) -> Tuple[ValidateCode, Union[Exception, str]]:
        """
        Encode a JWT
        :claims incoming claims
        :secret secret
        :validity validity in seconds
        :algorithm algorithm
        :compression compression

        :return ValidateCode or None, exception or None, encoded token or None:
        """
        if validity is not None:
            claims['exp'] = int((datetime.now() + timedelta(seconds=int(validity))).timestamp())

        encoded_cookie = None
        try:
            encoded_cookie = jwt.encode(claims, secret, algorithm=algorithm)
        except Exception as e:
            return ValidateCode.INVALID, e

        try:
            if compression:
                compressed_cookie = gzip.compress(encoded_cookie)
                encoded_cookie = base64.urlsafe_b64encode(compressed_cookie)
            encoded_cookie = str(encoded_cookie, 'utf-8')
        except Exception as e:
            return ValidateCode.UNABLE_TO_COMPRESS, e

        return ValidateCode.VALID, encoded_cookie

    @staticmethod
    def decode(*, cookie: str, secret: str = '', verify: bool = True,
               compression: bool = False) -> Tuple[ValidateCode, Union[Exception, dict]]:
        """
        Decode and validate a JWT
        :cookie incoming cookie
        :secret secret
        :compression compression

        :return ValidateCode or None, exception or None, Decoded Token or None:
        """
        try:
            if compression:
                decoded_64 = base64.urlsafe_b64decode(cookie)
                uncompressed_cookie = gzip.decompress(decoded_64)
                cookie = uncompressed_cookie
        except Exception as e:
            traceback.print_exc()
            return ValidateCode.UNABLE_TO_DECOMPRESS, e

        algorithm = None
        try:
            algorithm = jwt.get_unverified_header(cookie).get('alg', None)
        except jwt.DecodeError as e:
            traceback.print_exc()
            return ValidateCode.UNPARSABLE_TOKEN, e

        if algorithm is None:
            return ValidateCode.UNSPECIFIED_ALG, None

        try:
            decoded_token = jwt.decode(cookie, secret, algorithms=[algorithm], verify=verify)
            return ValidateCode.VALID, decoded_token
        except Exception as e:
            traceback.print_exc()
            return ValidateCode.INVALID, e
